@@ -21,7 +21,7 @@ class BaseRecordForm(forms.ModelForm):
     """
 
     class Meta:
-        fields = ("form",)
+        fields = ("_form",)
 
     def __init__(
         self,
@@ -32,7 +32,7 @@ class BaseRecordForm(forms.ModelForm):
         **kwargs: Any,
     ) -> None:
         initial_data = {
-            **(instance.data if instance else {}),
+            **(instance._data if instance else {}),
             **(initial or {}),
         }
 
@@ -74,22 +74,8 @@ class BaseRecordForm(forms.ModelForm):
         Returns:
             instance: The Record model instance.
         """
-        record = cast("BaseRecord", super().save(commit=False))
-
-        # If nothing has changed, noop and return the record as-is.
-        if not self.has_changed():
-            return record
-
         # Update any changed attributes.
         for field_name in self.changed_data:
-            cleaned_value = self.cleaned_data[field_name]
+            setattr(self.instance, field_name, self.cleaned_data[field_name])
 
-            if field_name in self.Meta.fields:
-                setattr(record, field_name, cleaned_value)
-            else:
-                record.set_attribute(field_name, cleaned_value, commit=False)
-
-        if commit:
-            record.save()
-
-        return record
+        return cast("BaseRecord", super().save(commit=commit))
